"""
Generalized Linear Model reserve estimators.
"""
import numpy as np
import pandas as pd 
import statsmodels.api as sm
import statsmodels.formula.api as smf


class GLMEstimator:
	
	def __init__(self, tri):

		pass


class GLMEstimatorResult:
	
	def __init__(self, **kwargs):
		
		pass
